--[[
Author: Starinnia
ChimaeronHealthPoint is an addon designed to give healers a quick list of raiders under 10k health
during the Chimaeron encounter in Blackwing Descent
Date: 2011-01-27T00:11:46Z
Revision: 14 
Project Version: 1.0.0 beta2
contact: codemaster2010 AT gmail DOT com

Copyright (c) 2011 Michael J. Murray aka Lyte of Lothar(US)
All rights reserved unless otherwise explicitly stated.
Thanks to Maat on WoWAce whose Atramedes addon was the inspiration for this
]]

local addon = LibStub("AceAddon-3.0"):NewAddon("ChimaeronHealthPoint", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("ChimaeronHealthPoint")

--upvalue globals used in health/ui updates
local pairs = pairs
local wipe = wipe
local tconcat = table.concat

local CHIMAERON = 43296
local resetString = L["The Bile-O-Tron springs to life and begins to emit a foul smelling substance."]
local unlock = "Interface\\AddOns\\ChimaeronHealthPoint\\Textures\\un_lock"
local lock = "Interface\\AddOns\\ChimaeronHealthPoint\\Textures\\lock"
local forced = false
local threshold = 10000

--tables for displaying the info
local lowHealthPlayers = {}
local sortedHealths = {}
local stringBuilder = {}

--save the hex color escape codes
local hexColors = {}
for k, v in pairs(RAID_CLASS_COLORS) do
	hexColors[k] = format("|cff%02x%02x%02x", v.r * 255, v.g * 255, v.b * 255)
end

local function getID(guid)
	return tonumber(strsub(guid,7,10),16)
end

local function short(value)
	return format("%.1fk", value / 1000)
end

--forward declaration of the functions for the lock functions
local lockDisplay
local unlockDisplay
local updateLockButton
local toggleLock

function addon:OnInitialize()
	local defaults = {
		profile = {
			disabled = false,
			position = {},
			locked = false,
			width = 100,
			height = 122,
			maxDisplay = 10,
		},
	}
	self.db = LibStub("AceDB-3.0"):New("ChimaeronHealthPointDB", defaults, "Default")
	
	_G["SlashCmdList"]["CHIMAERONHEALTHPOINT_MAIN"] = function(s)
		local cmd, amt = strsplit(" ", s)
		cmd = string.lower(cmd)
		amt = tonumber(amt) or -1
		if cmd == "disable" then
			if self.db.profile.disabled then
				print(format("|cff33ff99ChimaeronHealthPoint|r: %s", L["Already disabled!"]))
			else
				self:OnDisable()
				self.db.profile.disabled = true
				print(format("|cff33ff99ChimaeronHealthPoint|r: |cffff0000%s|r", L["Disabled"]))
			end
		elseif cmd == "enable" then
			if self.db.profile.disabled then
				self.db.profile.disabled = false
				self:OnEnable()
				print(format("|cff33ff99ChimaeronHealthPoint|r: |cff00ff00%s|r", L["Enabled"]))
			else
				print(format("|cff33ff99ChimaeronHealthPoint|r: %s", L["Already enabled!"]))
			end
		elseif cmd == "lock" then
			toggleLock()
			local suffix = self.db.profile.locked and L["Locked"] or L["Unlocked"]
			local color = self.db.profile.locked and "ff0000" or "00ff00"
			print(format("|cff33ff99ChimaeronHealthPoint|r: |cff%s%s|r", color, suffix))
		elseif cmd == "limit" then
			if amt > 25 then amt = 25 end
			if amt < 0 then amt = 1 end
			self.db.profile.maxDisplay = amt
			print(format("|cff33ff99ChimaeronHealthPoint|r: %s", format(L["Display limit set to %d players"], amt)))
		elseif cmd == "force" then
			forced = true
			self:CreateUI()
			self.ui.header:SetFormattedText("HP < %s", short(amt))
			threshold = amt
			self:RegisterEvent("UNIT_HEALTH")
			self.ui:Show()
			print(format("|cff33ff99ChimaeronHealthPoint|r: %s", format(L["Health threshold set to %s"], short(amt))))
			print(format("|cff33ff99ChimaeronHealthPoint|r: %s", L["Closing the display will reset the threshold to 10k, and return the addon to whatever state it was in before being forced."]))
		elseif cmd == "standby" then
			self:UnregisterEvent("UNIT_HEALTH")
			self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
			self:UnregisterEvent("CHAT_MSG_MONSTER_EMOTE")
			self.ui:Hide()
			print(format("|cff33ff99ChimaeronHealthPoint|r: %s", L["Standby Mode Enabled"]))
		else
			if self.ui:IsVisible() then
				self.ui:Hide()
				if forced then
					self:UnregisterEvent("UNIT_HEALTH")
					self.ui.header:SetText("HP < 10.0k")
					threshold = 10000
					wipe(lowHealthPlayers)
					wipe(sortedHealths)
					wipe(stringBuilder)
					forced = false
				end
			else
				self.ui:Show()
			end
		end
	end
	_G["SLASH_CHIMAERONHEALTHPOINT_MAIN1"] = "/chimaeron"
	_G["SLASH_CHIMAERONHEALTHPOINT_MAIN2"] = "/chp"
end

function addon:OnEnable()
	if not self.db.profile.disabled then
		self:CreateUI()
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	end
end

function addon:OnDisable()
	self:UnregisterAllEvents()
	self.ui:Hide()
end

function addon:CheckEnable(unit)
	if UnitExists(unit) and getID(UnitGUID(unit)) == CHIMAERON then
		self:RegisterEvent("UNIT_HEALTH")
		self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

function addon:PLAYER_TARGET_CHANGED()
	self:CheckEnable("target")
end

function addon:UPDATE_MOUSEOVER_UNIT()
	self:CheckEnable("mouseover")
end

function addon:CHAT_MSG_MONSTER_EMOTE(event, msg)
	if msg == resetString then
		--reset on new fight
		wipe(lowHealthPlayers)
		self.ui:Show()
		self:UpdateUI()
	end
end

function addon:UNIT_HEALTH(event, unit)
	if not UnitInRaid(unit) then return end
	
	local hp = UnitHealth(unit)
	if hp < threshold then
		lowHealthPlayers[UnitName(unit)] = hp
	else
		lowHealthPlayers[UnitName(unit)] = nil
	end
	
	self:UpdateUI()
end

function addon:COMBAT_LOG_EVENT_UNFILTERED(_, _, subevent, _, _, _, dstGUID)
	if subevent == "UNIT_DIED" then
		if getID(dstGUID) == CHIMAERON then
			self:UnregisterAllEvents()
			wipe(lowHealthPlayers)
			wipe(sortedHealths)
			wipe(stringBuilder)
			self.ui:Hide()
		end 
	end
end

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local _, class = UnitClass(key)
		if class then
			self[key] = hexColors[class] .. key .. "|r"
			return self[key]
		else
			return key
		end
	end
})

local function sortFunc(a, b) return lowHealthPlayers[a] < lowHealthPlayers[b] end
function addon:UpdateUI()
	--order the names with lowest health first
	wipe(sortedHealths)
	for k in pairs(lowHealthPlayers) do
		sortedHealths[#sortedHealths+1] = k
	end
	table.sort(sortedHealths, sortFunc)
	
	--use a table as a string builder
	wipe(stringBuilder)
	for i = 1, #sortedHealths do
		local n = sortedHealths[i]
		if n and UnitExists(n) and not UnitIsDeadOrGhost(n) then
			stringBuilder[#stringBuilder + 1] = coloredNames[n]
		end
		--end early if we hit the user defined limit
		if i > self.db.profile.maxDisplay then break end
	end
	
	--update the text on the display
	if #stringBuilder == 0 then
		self.ui.text:SetText("|cff777777:-P|r")
	else
		self.ui.text:SetText(tconcat(stringBuilder, "\n"))
	end
end

local function onDragStart(self) self:StartMoving() end
local function onDragStop(self)
	self:StopMovingOrSizing()
	local point, _, anchor, x, y = self:GetPoint()
	addon.db.profile.position.x = floor(x)
	addon.db.profile.position.y = floor(y)
	addon.db.profile.position.anchor = anchor
	addon.db.profile.position.point = point
end
local function OnDragHandleMouseDown(self) self.frame:StartSizing("BOTTOMRIGHT") end
local function OnDragHandleMouseUp(self) self.frame:StopMovingOrSizing() end
local function onResize(self, width, height)
	addon.db.profile.width = width
	addon.db.profile.height = height
end

local function lockDisplay()
	addon.ui:EnableMouse(false)
	addon.ui:SetMovable(false)
	addon.ui:SetResizable(false)
	addon.ui:RegisterForDrag()
	addon.ui:SetScript("OnSizeChanged", nil)
	addon.ui:SetScript("OnDragStart", nil)
	addon.ui:SetScript("OnDragStop", nil)
	addon.ui.drag:Hide()
end

local function unlockDisplay()
	addon.ui:EnableMouse(true)
	addon.ui:SetMovable(true)
	addon.ui:SetResizable(true)
	addon.ui:RegisterForDrag("LeftButton")
	addon.ui:SetScript("OnSizeChanged", onResize)
	addon.ui:SetScript("OnDragStart", onDragStart)
	addon.ui:SetScript("OnDragStop", onDragStop)
	addon.ui.drag:Show()
end

local function updateLockButton()
	if not addon.ui then return end
	addon.ui.lock:SetNormalTexture(addon.db.profile.locked and lock or unlock)
end

local function toggleLock()
	addon.db.profile.locked = not addon.db.profile.locked
	if addon.db.profile.locked then
		lockDisplay()
	else
		unlockDisplay()
	end
	updateLockButton()
end

local function onControlEnter(self)
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:AddLine(self.tooltipHeader)
	GameTooltip:AddLine(self.tooltipText, 1, 1, 1, 1)
	GameTooltip:Show()
end
local function onControlLeave() GameTooltip:Hide() end
local function closeWindow()
	addon.ui:Hide()
	if forced then
		addon:UnregisterEvent("UNIT_HEALTH")
		addon.ui.header:SetText("HP < 10.0k")
		threshold = 10000
		wipe(lowHealthPlayers)
		wipe(sortedHealths)
		wipe(stringBuilder)
		forced = false
	end
end

function addon:CreateUI()
	if self.ui then return end
	
	local f = CreateFrame("FRAME", nil, UIParent)
	f:SetWidth(self.db.profile.width)
	f:SetHeight(self.db.profile.height)
	f:SetClampedToScreen(true)
	f:SetMinResize(100, 60)
	
	f.bg = f:CreateTexture(nil, "PARENT")
	f.bg:SetAllPoints(f)
	f.bg:SetBlendMode("BLEND")
	f.bg:SetTexture(0, 0, 0, 0.5)
	
	if self.db.profile.position.x then
		f:SetPoint(self.db.profile.position.point, UIParent, self.db.profile.position.anchor, self.db.profile.position.x, self.db.profile.position.y)
	else
		f:SetPoint("CENTER")
	end
	
	f:SetScript("OnDragStart", onDragStart)
	f:SetScript("OnDragStop", onDragStop)
	f:SetScript("OnSizeChanged", onResize)
	
	f.close = CreateFrame("Button", nil, f)
	f.close:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT", -2, 2)
	f.close:SetHeight(16)
	f.close:SetWidth(16)
	f.close.tooltipHeader = L["Close"]
	f.close.tooltipText = L["Closes the Health Threshold display."]
	f.close:SetNormalTexture("Interface\\AddOns\\ChimaeronHealthPoint\\Textures\\close")
	f.close:SetScript("OnEnter", onControlEnter)
	f.close:SetScript("OnLeave", onControlLeave)
	f.close:SetScript("OnClick", closeWindow)

	f.lock = CreateFrame("Button", nil, f)
	f.lock:SetPoint("BOTTOMLEFT", f, "TOPLEFT", 2, 2)
	f.lock:SetHeight(16)
	f.lock:SetWidth(16)
	f.lock.tooltipHeader = L["Toggle lock"]
	f.lock.tooltipText = L["Toggle whether or not the window should be locked or not."]
	f.lock:SetScript("OnEnter", onControlEnter)
	f.lock:SetScript("OnLeave", onControlLeave)
	f.lock:SetScript("OnClick", toggleLock)
	
	f.header = f:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	f.header:SetText("HP < 10.0k")
	f.header:SetPoint("BOTTOM", f, "TOP", 0, 4)
	
	f.text = f:CreateFontString(nil, "OVERLAY", "GameFontNormalHuge")
	f.text:SetFont("Fonts\\FRIZQT__.TTF", 12)
	f.text:SetText("")
	f.text:SetAllPoints(f)
	f:SetScript("OnShow", function(frame) frame.text:SetText("|cff777777:-P|r") end)
	
	f.drag = CreateFrame("Frame", nil, f)
	f.drag.frame = f
	f.drag:SetFrameLevel(f:GetFrameLevel() + 10)
	f.drag:SetWidth(16)
	f.drag:SetHeight(16)
	f.drag:SetPoint("BOTTOMRIGHT", f, -1, 1)
	f.drag:EnableMouse(true)
	f.drag:SetScript("OnMouseDown", OnDragHandleMouseDown)
	f.drag:SetScript("OnMouseUp", OnDragHandleMouseUp)
	f.drag:SetAlpha(0.5)
	
	f.drag.tex = f.drag:CreateTexture(nil, "BACKGROUND")
	f.drag.tex:SetTexture("Interface\\AddOns\\ChimaeronHealthPoint\\Textures\\draghandle")
	f.drag.tex:SetWidth(16)
	f.drag.tex:SetHeight(16)
	f.drag.tex:SetBlendMode("ADD")
	f.drag.tex:SetPoint("CENTER", f.drag)
	
	f:Hide()
	self.ui = f
	
	if self.db.profile.locked then
		lockDisplay()
	else
		unlockDisplay()
	end
	updateLockButton()
end
